//Microsoft Developer Studio generated resource script.
//
#include "resource.h"

#define APSTUDIO_READONLY_SYMBOLS
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 2 resource.
//
#include "newres.h"

/////////////////////////////////////////////////////////////////////////////
#undef APSTUDIO_READONLY_SYMBOLS

/////////////////////////////////////////////////////////////////////////////
// English (U.S.) resources

#if !defined(AFX_RESOURCE_DLL) || defined(AFX_TARG_ENU)
#ifdef _WIN32
LANGUAGE LANG_ENGLISH, SUBLANG_ENGLISH_US
#pragma code_page(1252)
#endif //_WIN32

/////////////////////////////////////////////////////////////////////////////
//
// Icon
//

// Icon with lowest ID value placed first to ensure application icon
// remains consistent on all systems.
IDI_NIOUP               ICON    DISCARDABLE     "vlc.ico"

#ifdef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// TEXTINCLUDE
//

1 TEXTINCLUDE DISCARDABLE 
BEGIN
    "resource.h\0"
END

2 TEXTINCLUDE DISCARDABLE 
BEGIN
    "#include ""newres.h""\r\n"
    "\0"
END

3 TEXTINCLUDE DISCARDABLE 
BEGIN
    "\r\n"
    "\0"
END

#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Menu
//

IDM_MENU MENU DISCARDABLE 
BEGIN
    POPUP "&File"
    BEGIN
        MENUITEM "Open File...",                ID_FILE_OPENFILE, GRAYED
        MENUITEM "Network Stream...",           ID_FILE_NETWORKSTREAM
        , GRAYED
        MENUITEM SEPARATOR
        MENUITEM "E&xit",                       IDM_FILE_EXIT
    END
    POPUP "View"
    BEGIN
        MENUITEM "Playlist...",                 ID_VIEW_PLAYLIST, GRAYED
        MENUITEM "Messages...",                 ID_VIEW_MESSAGES, GRAYED
    END
    POPUP "Settings"
    BEGIN
        MENUITEM "Audio",                       ID_SETTINGS_AUDIO, GRAYED
        MENUITEM "Subtitles",                   ID_SETTINGS_SUBTITLES
        , GRAYED
        MENUITEM SEPARATOR
        MENUITEM "Preferences...",              ID_SETTINGS, GRAYED
    END
    POPUP "&Help"
    BEGIN
        MENUITEM "&About...",                   IDM_HELP_ABOUT
    END
END


/////////////////////////////////////////////////////////////////////////////
//
// Dialog
//

IDD_ABOUTBOX DIALOG DISCARDABLE  0, 0, 127, 42
STYLE DS_MODALFRAME | WS_POPUP | WS_CAPTION | WS_SYSMENU
EXSTYLE 0x80000000L
CAPTION "About PocketVLC"
FONT 8, "System"
BEGIN
    ICON            IDI_NIOUP,IDC_STATIC,7,7,21,20
    LTEXT           "VLC Version Tarass",IDC_STATIC,38,10,70,8,SS_NOPREFIX
    LTEXT           "Copyright (C) VideoLAN (Centrale Reseaux) and its contributors",IDC_STATIC,30,27,80,8
END


/////////////////////////////////////////////////////////////////////////////
//
// DESIGNINFO
//

#ifdef APSTUDIO_INVOKED
GUIDELINES DESIGNINFO DISCARDABLE 
BEGIN
    IDD_ABOUTBOX, DIALOG
    BEGIN
        LEFTMARGIN, 7
        RIGHTMARGIN, 120
        TOPMARGIN, 7
        BOTTOMMARGIN, 35
    END
END
#endif    // APSTUDIO_INVOKED


/////////////////////////////////////////////////////////////////////////////
//
// Accelerator
//

IDC_NIOUP ACCELERATORS DISCARDABLE 
BEGIN
    VK_F1,          IDM_PLOP,               VIRTKEY, NOINVERT
    VK_F4,          IDM_FILE_EXIT,          VIRTKEY, ALT, NOINVERT
    VK_F8,          IDM_HELP_ABOUT,         VIRTKEY, NOINVERT
END

#endif    // English (U.S.) resources
/////////////////////////////////////////////////////////////////////////////



#ifndef APSTUDIO_INVOKED
/////////////////////////////////////////////////////////////////////////////
//
// Generated from the TEXTINCLUDE 3 resource.
//


/////////////////////////////////////////////////////////////////////////////
#endif    // not APSTUDIO_INVOKED

