/*****************************************************************************
 * style.css: VLC media player web interface
 *****************************************************************************
 * Copyright (C) 2005-2006 the VideoLAN team
 * $Id$
 *
 * Authors: Antoine Cellerier <dionoea -at- videolan -dot- org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

body {
	background-color: #eee;
	font-family: Arial, sans-serif;
	font-size: 12pt;
	text-align: center; /* Center on page - IE */
	padding: 0px;
	margin: 0px;
}

div {
	padding: 0px;
	margin: 0px;
	text-align: left;
}

div.dialog {
	width: 440pt;
	background: #fff;
	border: solid #000 1px;
	margin: 10px auto; /* Center on page - Firefox */
	padding: 0px;
	overflow: hidden; /* so that the title bar doesn't overflow on
                           * firefox but still takes all the div's width
                           * in IE */
}

div#footer {
	font-size: 8pt;
	text-align: center;
}

.btn_text {
	display: none;
}

form {
	display: inline;
}

input, textarea {
	border: solid #000 1px;
	background-color: #fff;
}

div.helper {
	margin: 10px;
	padding: 10px 0px;
	border: solid 1px #000;
}
div.helper hr {
	border: solid 1px #000;
}

div.title {
	width: 100%/*576px*/; /* see overflow:hidden several lines 
                                * before for explanation */
	background: #000 url('/images/vlc16x16.png') no-repeat top left;
	padding-left: 24px;
	color: #fff;
	font-weight: bold;
}
div.title button {
	border: 1px none #000;
	padding: 0px;
	background-color: #000;
	color: #fff;
}

div.controls {
	width: 100%;
	padding: 3px 5px;
}
div.controls button {
	border: 1px solid #fff;
	padding: 0px;
	background-color: #fff;
}

div.list {
	padding: 1em;
}
div.list_element {
	padding-bottom: 0.3em;
}
div.list_element ul {
	margin: 0px;
}

div#infotree ul {
	padding: 0.4em;
	margin: 0em;
}
div#infotree li {
	font-weight: bold;
	font-size: 0.8em;
}
div#infotree dl {
	font-weight: normal;
	padding: 0em 1em;
}
div#infotree dt {
	text-decoration: underline;
}

div#playtree {
	min-height: 150px;
}
div.pl_node {
	padding-left: 20px;
	font-style: italic;
}
a.pl_leaf {
	font-style: normal;
}
a.pl_leaf:hover {
	color: #f00;
}
img {
	border: 0px none black;
}

div.popup {
	background-color: #fff;
	overflow: hidden;
	border: solid #888 1px;
	margin: 10px auto; /* Center on page - Firefox */
	position: absolute;
	z-index: 2;
	font-size: 0.8em;
}
div#browse {
	width: 70%;
	left: 15%;
}
div#browse div.title {
	background-color: #008;
}

div#browser {
	padding: 20px;
}

div#mosaic_list {
	width: 50%;
	left: 25%;
	background: url('/old/images/white.png') repeat;
	display: none;
}

.mosaic_bg {
	background-color: #aaf;
}
.mosaic_tbl {
	background-color: #faa;
}
.mosaic_itm {
	background-color: #afa;
}
div#mosaic_layout {
	border: 0px solid #000;
	padding: 0px;
	margin: 10px auto;
}

div#mosaic_dt {
	border: 0px solid #000;
	position: relative;
}
table#mosaic_table {
	border-spacing: 0px 0px;
	position: relative;
}
table#mosaic_table tr {
	padding: 0px;
	margin: 0px;
}
table#mosaic_table td {
	border: 0px solid #000;
	text-align: center;
	padding: 0px;
	margin: 0px;
}
table#mosaic_table input {
	border: 0px;
	background: transparent;
	text-align: center;
}
textarea#mosaic_code {
	padding: 0px;
	display: block;
	margin: 10px auto;
}

div.controls button.on {
	background-color: #aaa;
}

div.controls button.off {
	background-color: #fff;
}

div.menubtn {
	display: inline;
}

div.menu {
	position: absolute;
	border: 1px solid #ccc;
	padding: 5px;
	margin-top: -5px;
	background: url('/old/images/white.png') repeat;
	z-index: 2;
	display: none;
}

/* FIXME: still some div.menu > button around ? */
div.menu div,
div.menu button {
    background: transparent;
    cursor: pointer;
    white-space: nowrap;
}

div.menu button.menuout {
	border: 1px solid transparent;
}

div.menu div.menuover,
div.menu button.menuover {
    background: #000;
    color: #fff;
}

div.popup img.close {
	position: absolute;
	right: 2px;
	top: 2px;
}
