/*****************************************************************************
 * Copyright © 2011 VideoLAN
 * $Id: 51f28fb643c6136d4fb59f885fd1d1ae9a6818b5 $
 *
 * Authors: Filipe Azevedo, aka PasNox
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#ifndef ROUNDBUTTON_H
#define ROUNDBUTTON_H

#include <QToolButton>

class RoundButton : public QToolButton
{
    Q_OBJECT

public:
    RoundButton( QWidget* parent = 0 );

    virtual QSize sizeHint() const;

protected:
    QBrush pen( QStyleOptionToolButton* option ) const;
    QBrush brush( QStyleOptionToolButton* option ) const;

    virtual void paintEvent( QPaintEvent* event );
};

#endif // ROUNDBUTTON_H
