/*****************************************************************************
 * id3genres.h: list of genres for id3 genre tags
 *****************************************************************************
 * Copyright (C) 2002-2003 the VideoLAN team
 * $Id: 4f9d648e060333c9b7e9b1f6cc900db6af92d3b9 $
 *
 * Author: Sigmund Augdal Helberg <dnumgis@videolan.org>
 *         list found at http://www.id3.org/id3v2.4.0-frames.txt
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#define NUM_GENRES 80

static const char *ppsz_genres[] = {
    N_("Blues"),
    N_("Classic rock"),
    N_("Country"),
    N_("Dance"),
    N_("Disco"),
    N_("Funk"),
    N_("Grunge"),
    N_("Hip-Hop"),
    N_("Jazz"),
    N_("Metal"),
    N_("New Age"),
    N_("Oldies"),
    N_("Other"),
    N_("Pop"),
    N_("R&B"),
    N_("Rap"),
    N_("Reggae"),
    N_("Rock"),
    N_("Techno"),
    N_("Industrial"),
    N_("Alternative"),
    N_("Ska"),
    N_("Death metal"),
    N_("Pranks"),
    N_("Soundtrack"),
    N_("Euro-Techno"),
    N_("Ambient"),
    N_("Trip-Hop"),
    N_("Vocal"),
    N_("Jazz+Funk"),
    N_("Fusion"),
    N_("Trance"),
    N_("Classical"),
    N_("Instrumental"),
    N_("Acid"),
    N_("House"),
    N_("Game"),
    N_("Sound clip"),
    N_("Gospel"),
    N_("Noise"),
    N_("Alternative rock"),
    N_("Bass"),
    N_("Soul"),
    N_("Punk"),
    N_("Space"),
    N_("Meditative"),
    N_("Instrumental pop"),
    N_("Instrumental rock"),
    N_("Ethnic"),
    N_("Gothic"),
    N_("Darkwave"),
    N_("Techno-Industrial"),
    N_("Electronic"),
    N_("Pop-Folk"),
    N_("Eurodance"),
    N_("Dream"),
    N_("Southern rock"),
    N_("Comedy"),
    N_("Cult"),
    N_("Gangsta"),
    N_("Top 40"),
    N_("Christian rap"),
    N_("Pop/funk"),
    N_("Jungle"),
    N_("Native American"),
    N_("Cabaret"),
    N_("New wave"),
    N_("Psychedelic"),
    N_("Rave"),
    N_("Showtunes"),
    N_("Trailer"),
    N_("Lo-Fi"),
    N_("Tribal"),
    N_("Acid punk"),
    N_("Acid jazz"),
    N_("Polka"),
    N_("Retro"),
    N_("Musical"),
    N_("Rock & roll"),
    N_("Hard rock")
};
