/*****************************************************************************
 * maemo_interface.h : Interface creation header file for the maemo plugin.
 *****************************************************************************
 * Copyright (C) 2008 the VideoLAN team
 * $Id: 319b14e6dde1a86bda6cc4f4b513a0095828ebb4 $
 *
 * Authors: Antoine Lejeune <phytos@videolan.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston MA 02110-1301, USA.
 *****************************************************************************/

#include <gtk/gtk.h>

#include <vlc_common.h>

void create_playlist( intf_thread_t *p_intf );
void delete_playlist( intf_thread_t *p_intf );
